import pytest
from scalib.metrics import Ttest
import numpy as np


def reference(traces, x, D):
    CM0 = np.zeros((D * 2, traces.shape[1]))
    I0 = np.where(x == 0)[0]
    u0 = np.mean(traces[I0, :], axis=0)
    for d in range(D * 2):
        CM0[d, :] = np.mean((traces[I0, :] - u0) ** (d + 1), axis=0)

    CM1 = np.zeros((D * 2, traces.shape[1]))
    I1 = np.where(x == 1)[0]
    u1 = np.mean(traces[I1, :], axis=0)
    for d in range(D * 2):
        CM1[d, :] = np.mean((traces[I1, :] - u1) ** (d + 1), axis=0)
    u1_ref = u1
    u0_ref = u0
    t = np.zeros((D, len(traces[0, :])))
    n = [len(I0), len(I1)]
    for d in range(1, D + 1):
        if d == 1:
            u0 = u0_ref
            u1 = u1_ref
            v0 = CM0[1, :]
            v1 = CM1[1, :]
        elif d == 2:
            u0 = CM0[1, :]
            u1 = CM1[1, :]
            v0 = CM0[3, :] - CM0[1, :] ** 2
            v1 = CM1[3, :] - CM1[1, :] ** 2
        else:
            u0 = CM0[d - 1, :] / np.power(CM0[1, :], d / 2)
            u1 = CM1[d - 1, :] / np.power(CM1[1, :], d / 2)
            v0 = (CM0[(d * 2) - 1, :] - CM0[(d) - 1, :] ** 2) / (CM0[1, :] ** d)
            v1 = (CM1[(d * 2) - 1, :] - CM1[(d) - 1, :] ** 2) / (CM1[1, :] ** d)

        t[d - 1, :] = (u0 - u1) / (np.sqrt((v0 / n[0]) + (v1 / n[1])))
    return t


def test_ttest_d1():
    ns = 100
    d = 1
    nc = 2
    n = 200

    m = np.random.randint(0, 2, (nc, ns))
    traces = np.random.randint(0, 10, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    traces += m[labels]

    t_ref = reference(traces, labels, d)
    ttest = Ttest(1)
    ttest.fit_u(traces, labels)
    t = ttest.get_ttest()
    assert np.allclose(t_ref, t, rtol=1e-3)


def test_ttest_d2():
    ns = 100
    d = 2
    nc = 2
    n = 200

    m = np.random.randint(0, 2, (nc, ns))
    traces = np.random.randint(0, 10, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    traces += m[labels]

    t_ref = reference(traces, labels, d)
    ttest = Ttest(d)
    ttest.fit_u(traces, labels)
    t = ttest.get_ttest()
    assert np.allclose(t_ref, t, rtol=1e-3)


def test_ttest_d6_multiple_fit():
    ns = 12300
    d = 6
    nc = 2
    n = 200

    m = np.random.randint(0, 2, (nc, ns))
    traces = np.random.randint(0, 10, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    traces += m[labels]

    t_ref = reference(traces, labels, d)
    ttest = Ttest(d)
    for i in range(0, n, 10):
        ttest.fit_u(traces[i : i + 10, :], labels[i : i + 10])
    t = ttest.get_ttest()
    assert np.allclose(t_ref, t, rtol=1e-3)


def test_ttest_d6():
    ns = 942
    d = 6
    nc = 2
    n = 200

    m = np.random.randint(0, 2, (nc, ns))
    traces = np.random.randint(0, 10, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    traces += m[labels]

    t_ref = reference(traces, labels, d)
    ttest = Ttest(d)
    ttest.fit_u(traces, labels)
    t = ttest.get_ttest()
    assert np.allclose(t_ref, t, rtol=1e-3)



r"""

Linear Discriminant Analysis (LDA)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

LDA is a also known as pooled Gaussian templates in the side-channel litterature.
:class:`LDAClassifier` is the main implementation of the LDA, and
:class:`MultiLDA` is a convenience wrapper for profiling multiple
variables using the same traces (but possibly different sets of POIs).

.. currentmodule:: scalib.modeling

.. autosummary::
   :toctree:
   :recursive:
   :nosignatures:

   LDAClassifier
   MultiLDA
   RLDAClassifier
"""

__all__ = ["LDAClassifier", "MultiLDA", "RLDAClassifier"]

from .ldaclassifier import LDAClassifier, MultiLDA
from .rldaclassifier import RLDAClassifier



import pytest
from scalib import ScalibError
from scalib.metrics import SNR
import numpy as np


def gen_snr_data(nv, nc, ns, n):
    ii16 = np.iinfo(np.int16)
    x = np.random.randint(ii16.min, ii16.max, (n, ns), dtype=np.int16)
    x[0, 0] = ii16.min
    if n > 1:
        x[1, 0] = ii16.max
    y = np.random.randint(0, nc, (n, nv), dtype=np.uint16)
    return x, y


def snr_pooled(x, y, nv, nc, ns):
    snr_ref = np.zeros((nv, ns), dtype=np.float64)
    g_mean = np.mean(x, axis=0)
    for v in range(nv):
        sig_acc = np.zeros((ns,))
        noise_acc = np.zeros((ns,))
        for c in range(nc):
            samples = x[np.where(y.transpose()[v] == c)[0], :]
            if samples.shape[0] != 0:
                mean = np.mean(samples, axis=0)
                sig_acc += (mean - g_mean) ** 2 * samples.shape[0]
                noise_acc += ((samples - mean[np.newaxis, :]) ** 2).sum(axis=0)
        with np.errstate(divide="ignore", invalid="ignore"):
            snr_ref[v, :] = (sig_acc / x.shape[0]) / (noise_acc / x.shape[0])
    return snr_ref


def snr_pooled2(x, y, nv, nc, ns, use_64bit):
    """Should be almost exact implementation of rust version, including rounding
    errors."""

    if use_64bit:
        stype = np.int64
    else:
        stype = np.int32

    sig = np.zeros((nv, ns), dtype=np.float64)
    no = np.zeros((nv, ns), dtype=np.float64)
    g_sum = np.sum(x.astype(np.int64), axis=0).astype(object)
    sum_sq = np.sum(x.astype(np.int64) ** 2, axis=0).astype(object)

    n = x.shape[0]
    for v in range(nv):
        sum_sq_cls = np.zeros((ns,)).astype(object)
        for c in range(nc):
            samples = x[np.where(y.transpose()[v] == c)[0], :]
            if samples.shape[0] != 0:
                s = np.sum(samples.astype(stype), axis=0).astype(object)
                sum_sq_cls += (s**2).astype(object) * n // samples.shape[0]
        sig[v, :] = (sum_sq_cls - g_sum**2).astype(np.float64)
        assert n == x.shape[0]
        no[v, :] = (x.shape[0] * sum_sq - sum_sq_cls).astype(np.float64)
    with np.errstate(divide="ignore", invalid="ignore"):
        snr_ref = sig / no
    return snr_ref


def snr_run(nc, nv, ns, n, use_64bit, batch=100):
    """
    Test the SNR.
    """
    print(f"SNR run {nc=} {nv=} {ns=} {n=} {use_64bit=} {batch=}")
    x, y = gen_snr_data(nv, nc, ns, n)

    # compute SNR with SCALib
    snr = SNR(nc=nc, use_64bit=use_64bit)

    n_samples = np.array([[(y[:, v] == c).sum() for c in range(nc)] for v in range(nv)])

    if (n_samples >= (2**16)).any() and not use_64bit:
        with pytest.raises(Exception):
            for i in range(0, n, batch):
                snr.fit_u(x[i : i + batch, :], y[i : i + batch, :])
            snr_val = snr.get_snr()
        return

    for i in range(0, n, batch):
        snr.fit_u(x[i : i + batch, :], y[i : i + batch, :])
    snr_val = snr.get_snr()

    snr_ref = snr_pooled2(x, y, nv, nc, ns, use_64bit)
    assert np.allclose(snr_ref, snr_val, equal_nan=True)
    snr_ref = snr_pooled(x, y, nv, nc, ns)
    if not (n_samples == 0).any():
        assert np.allclose(snr_ref, snr_val, equal_nan=True, rtol=1 / n, atol=0.1 / n)


def test_snr():
    test_cases = [
        {"nc": 2, "nv": 1, "ns": 1, "n": 1},
        {"nc": 2, "nv": 1, "ns": 1, "n": 4},
        {"nc": 2, "nv": 1, "ns": 10, "n": 4},
        {"nc": 2, "nv": 1, "ns": 1, "n": 10},
        {"nc": 2, "nv": 1, "ns": 1, "n": 1000},
        {"nc": 16, "nv": 5, "ns": 10, "n": 1000},
        {"nc": 16, "nv": 32, "ns": 10, "n": 10},
        {"nc": 16, "nv": 153, "ns": 10, "n": 10},
        {"nc": 1024, "nv": 3, "ns": 10, "n": 10},
        {"nc": 2, "nv": 3, "ns": 10**4, "n": 10},
        {"nc": 2, "nv": 1, "ns": 1, "n": 1 << 17},
    ]
    for test_case in test_cases:
        for use_64bit in [False, True]:
            snr_run(use_64bit=use_64bit, **test_case)


def test_large_snr():
    # Was deadlocking on error (only when displaying progress bar, i.e., for
    # large cases)
    ntraces = 1000
    ns = 10000
    nv = 1000
    traces = np.random.randint(0, 256, (ntraces, ns), dtype=np.int16)
    nb = 16
    X = np.random.randint(0, nb, (ntraces, nv), dtype=np.uint16)
    snr = SNR(nb)
    snr.fit_u(traces, X)
    snr.get_snr()

    snr = SNR(nb - 1)
    with pytest.raises(ScalibError):
        snr.fit_u(traces, X)


def test_extreme_nc():
    # Issue #54
    for nc in (1, 2**16 + 1):
        with pytest.raises(Exception):
            SNR(nc)
    for nc in (2, 2**16):
        SNR(nc)



import os
import sys

from setuptools import setup
from setuptools_rust import Binding, RustExtension

# Ensure these are present (in case we are not using PEP-518 compatible build
# system).
import setuptools_scm

def env_true(env):
    return env == '1'

noflags = env_true(os.environ.get("SCALIB_NOFLAGS"))
portable = not noflags and env_true(os.environ.get("SCALIB_PORTABLE"))
use_avx2 = not noflags and env_true(os.environ.get("SCALIB_AVX2"))

if portable and use_avx2:
    raise ValueError("Cannot have both SCALIB_PORTABLE and SCALIB_AVX2.")

# We check only for AVX2, as this is the CI default, otherwise we assume local
# builds.
with open("src/scalib/build_config.py", "w") as f:
    f.write(f"REQUIRE_AVX2 = {use_avx2}\n")


if noflags or portable:
    rustflags = None
elif use_avx2:
    rustflags = "-C target-feature=+avx2"
else:
    rustflags = "-C target-cpu=native"

if rustflags:
    rustflags = os.environ.get('RUSTFLAGS', '') + ' ' + rustflags
    os.environ["RUSTFLAGS"] = rustflags

print(f"Build config: {noflags=} {portable=} {use_avx2=} {rustflags=}.")

scalib_features = ["pyo3/abi3"]

if sys.platform == "linux":
    scalib_features.append("blis")

setup(
    project_urls={
        "Bug Tracker": "https://github.com/simple-crypto/scalib/issues",
    },
    rust_extensions=[
        RustExtension(
            "scalib._scalib_ext",
            path="src/scalib_ext/scalib-py/Cargo.toml",
            binding=Binding.PyO3,
            features=scalib_features,
            py_limited_api=True,
        )
    ],
)



import numpy as np
import numpy.typing as npt

from scalib import _scalib_ext
from scalib.config import get_config
import scalib.utils


class RLDAClassifier:
    r"""Regression-based Linear Discriminant Analysis.

    Models the leakage using a regression-based linear discriminant analysis
    (RLDA) classifier :footcite:p:`RLDA`, which can efficiently handle long
    traces and large number of classes.

    In a nutshell, this model performs LDA with the class means modelled as
    linear regression based on the :math:`n_b` bits of the class value.
    Compared to the :class:`scalib.modeling.LDAClassifier`, this model will
    perform better when the number of classes is large and/or there are few
    profiling traces.

    Internally, it first estimates the coefficients of the linear regression,
    then computes a projection matrix that reduces the dimensionality of the
    gaussian template to :math:`p` dimensions and makes the covariance matrix
    the identity.

    It is then able to predict the leakage likelihood

    .. math::
        \hat{\mathsf{f}}[\mathbf{l}|X=x] =
        \alpha
        \exp\left(
        -\frac{1}{2} \lVert\mathbf{W}^T\mathbf{l} - \mathbf{A}\mathbf\beta(x)\rVert^2
        \right).

    Where :math:`\mathbf{W}` is the projection matrix, :math:`\mathbf{A}` the projected
    regression coefficients, and :math:`\mathbf{\beta(x)}` the coefficients of :math:`x`.
    The parameter :math:`\alpha = 1/\sqrt{(2\pi)^p\lvert\hat\Sigma_\mathbf{W}}\rvert` does
    not need to be calculated as it will get canceled out when applying Bayes' law.

    :class:`RLDAClassifier` provides the probability for each of the :math:`2^{n_b}`
    classes with :meth:`predict_proba`.

    Examples
    --------

    >>> from scalib.modeling import RLDAClassifier
    >>> import numpy as np
    >>> traces_model = np.random.randint(0,256,(5000,10),dtype=np.int16)
    >>> labels_model = np.random.randint(0,256,(5000,1),dtype=np.uint64)
    >>> rlda = RLDAClassifier(8, 3)
    >>> rlda.fit_u(traces_model, labels_model)
    >>> rlda.solve()
    >>> traces_test = np.random.randint(0,256,(5000,10),dtype=np.int16)
    >>> prs = rlda.predict_proba(traces_test, 0)

    References
    ----------

    .. footbibliography::
    """

    def __init__(self, nb: int, p: int):
        """
        Parameters
        ----------
        nb:
            Number of bits of the profiled variables.
        nv:
            Number of variables to profile
        p:
            Number of dimensions in the linear subspace.
        """
        self._ns = None
        self._nv = None
        self._p = p
        self._nb = nb
        self._init = False
        self._solved = False

    def fit_u(
        self, traces: npt.NDArray[np.int16], x: npt.NDArray[np.uint64], gemm_mode=1
    ):
        """Update statistical model estimates with additional data.

        This can be called multiple times, the state is accumulated.

        Parameters
        ----------
        traces : array_like, int16
            Array that contains the traces. Shape `(n,ns)`.
        x : array_like, uint64
            Labels for each trace. Shape `(n,nv)`.
        """
        traces = scalib.utils.clean_traces(traces, self._ns)
        x = scalib.utils.clean_labels(x, self._nv, exp_type=np.uint64)
        if not self._init:
            self._init = True
            self._ns = traces.shape[1]
            self._nv = x.shape[1]
            self._inner = _scalib_ext.RLDA(self._nb, self._ns, self._nv, self._p)

        self._inner.update(traces, x.T, gemm_mode, get_config())

    def solve(self):
        """Solve the RLDA equations.

        Notes
        -----
        Once this has been called, predictions can be performed.
        """
        self._inner.solve(get_config())
        self._solved = True

    def get_proj(self) -> npt.NDArray[np.float64]:
        """Returns the projection matrix.

        Returns
        -------
        array_like, float64
            Shape (nv,p,ns)."""
        return self._inner.get_norm_proj()

    def get_proj_coefs(
        self,
    ) -> npt.NDArray[np.float64]:
        """The projected regression coefficients.

        Returns
        -------
        array_like, float64
            Shape (nv,p,nb+1).
        """
        return self._inner.get_proj_coefs()

    def predict_proba(
        self, traces: npt.NDArray[np.int16], var: int
    ) -> npt.NDArray[np.float64]:
        r"""Computes the probability for each of the classes for the requested variables.

        Parameters
        ----------
        traces:
            Array that contains the traces. Shape ``(n,ns)``.
        var:
            Id (position in the ``x`` array) of the variable for which the
            probabilities are computed.

        Returns
        -------
        array_like, f64
            Probabilities. Shape `(n, nc)`.
        """
        assert self._solved, "Model not solved"
        return self._inner.predict_proba(traces, var, get_config())

    class ClusteredModel:
        """Clustered RLDA model, see :func:`RLDAClassifier.get_clustered_model`."""

        pass

    def get_clustered_model(
        self,
        var: int,
        t: float,
        max_clusters: int = 10_000_000,
        store_associated_classes: bool = True,
    ) -> ClusteredModel:
        """Generate a simplified model for faster estimation of the information content in this model.

        This generates a model with clustered means that can be used to
        estimate the percevied or training information of the model. It applies
        a clustering method on the classes to regroup the closest ones up to a
        threshold distance :math:`t`.
        Internally, it uses a Kd-tree data structure to find the nearest cluster efficiently.
        Details on the clustering algorithm can be found in [1].

        The resulting model can be used with
        :class:`scalib.metrics.RLDAInformationEstimator` (see there for usage
        example).

        Parameters
        ----------
        var:
            Id (position in the ``x`` array) of the variable for which the
            probabilities are computed.
        t:
            Maximum distance between 2 cluster centers. This is a trade-off parameter between
            the tightness of the information bounds (lower value of t) and
            computation (time and memory) efficiency (higher value of t).
        max_clusters:
            The maximum number of clusters that can be generated. If during generation, this
            limit is exceeded, an exception is raised.
        store_associated_classes : bool
            If True, the generated model stores the classes associated to each cluster. This
            allows refining the information bounds by calculating using the exact class mean (and not the
            centroid it is associated to) for clusters that contribute the most to an untight bound.
            Note that this option requires significantly more RAM for high values of :math:`n_b`.

        Returns
        -------
        ClusteredModel
            A clustered model to be used in :class:`scalib.metrics.RLDAInformationEstimator`
        """
        res = self.ClusteredModel()
        res._inner = self._inner.get_clustered_model(
            var, store_associated_classes, t, max_clusters
        )
        return res



"""Configuration of SCALib's threadpool."""

import os

from scalib import _scalib_ext


class ThreadPool:
    """SCALib threadpool.

    All computationally-heavy operations of SCALib are run on a thread pool.
    To use the configured `ThreadPool`, use `thread_context` or
    `default_threadpool`.
    """

    def __init__(self, n_threads):
        self.n_threads = n_threads
        self._pool = None

    @property
    def pool(self):
        # We initialize the true threadpool lazily for 2 reasons:
        # - It avoids creating threads when importing SCALib, which reduces
        # import time and, more importantly, prevents bugs with the usage of
        # subprocesses (such as with concurrent.futures.ProcessPoolExecutor),
        # since the POSIX API does not allow fork'ing after threadds have been
        # created.
        # - It generally improves performance when the threadpool is not actually used.
        if self._pool is None:
            self._pool = _scalib_ext.ThreadPool(self.n_threads)
        return self._pool


def _default_num_threads():
    num_threads = os.environ.get("SCALIB_NUM_THREADS")
    if num_threads is None:
        num_threads = _scalib_ext.usable_parallelism()
    else:
        try:
            num_threads = int(num_threads)
        except ValueError:
            raise ValueError(
                "Environment variable SCALIB_NUM_THREADS must be an integer."
            )
    return num_threads



r"""
This module contains several well-know side-channel metrics.

.. currentmodule:: scalib.metrics

.. autosummary::
   :toctree:
   :recursive:
   :nosignatures:

   SNR
   ttest
   RLDAInformationEstimator
"""
__all__ = ["SNR", "ttest", "RLDAInformationEstimator"]

from .snr import SNR
from .ttest import Ttest
from .ttest import MTtest
from .information import RLDAInformationEstimator



import numpy as np
from scipy.linalg import eigh
from scalib.modeling import RLDAClassifier
from scalib.metrics import RLDAInformationEstimator
import pytest
from scalib import ScalibError


def label2bits(label, nbits):
    """Return the array of -1/1 coefficients corresponding to a label."""
    return [1] + [-1 if (label >> i) & 1 == 0 else 1 for i in range(nbits)]


def get_bits(data, nbits):
    assert data.ndim == 1
    return np.array([label2bits(int(x), nbits) for x in data], dtype=np.int8)


def rlda_inner_test(ns, n_components, nb, nv):
    nc = 2**nb
    n_model = 2000
    n_test = 500
    noise = 2000
    assert nv <= ns, "Test case error, use higher ns"
    # Generate profiling data
    m = np.ones((nv, nc, ns), dtype=np.int16)
    for v in range(nv):
        m[v, :, v] = (get_bits(np.arange(nc), nbits=nb).astype(np.int64) + 1).sum(
            axis=1
        ) * (v + 1)
    traces = np.random.randint(-noise, noise, (n_model, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, (nv, n_model), dtype=np.uint64)
    for v in range(nv):
        traces += m[v, labels[v]].astype(np.int16)
    # Generate test data
    test_traces = np.random.randint(-noise, noise, (n_test, ns), dtype=np.int16)
    test_labels = np.random.randint(0, nc, (nv, n_test), dtype=np.uint16)
    for v in range(nv):
        test_traces += m[v, test_labels[v]]

    ## Run SCALib RLDA and compare projection matrices
    rlda = RLDAClassifier(nb, n_components)
    rlda.fit_u(traces, labels.T, 1)
    rlda.solve()

    for v in range(nv):
        ## Start computing RLDA in python
        label_bits = get_bits(labels[v], nbits=nb).astype(np.int64)

        B = label_bits.T @ label_bits
        C = label_bits.T @ traces
        S_L = traces.astype(np.int64).T @ traces.astype(np.int64)

        A = np.linalg.solve(B, C)
        mu = C[0] / n_model
        SB = A.T @ B @ A - n_model * np.outer(mu, mu)
        SW = S_L + A.T @ B @ A - C.T @ A - A.T @ C

        ev, W = eigh(SB, SW, subset_by_index=(ns - n_components, ns - 1))
        W = W.T
        cov = W @ SW @ W.T * 1 / n_model

        evals, evecs = np.linalg.eigh(cov)
        Wnorm = evecs * (evals[:, np.newaxis] ** -0.5)

        Weff = W.T @ Wnorm
        Aeff = A @ Weff

        prs_ref = (test_traces @ Weff)[:, :, np.newaxis] - (
            get_bits(np.arange(2**nb, dtype=np.uint64), nbits=nb) @ Aeff
        ).T[np.newaxis]
        prs_ref = np.exp(-0.5 * (prs_ref**2).sum(axis=1))
        prs_ref = prs_ref / prs_ref.sum(axis=1, keepdims=True)

        prs = rlda.predict_proba(test_traces, v)

        assert np.allclose(prs, prs_ref)


def test_rlda():
    ns = 6
    n_components = 3
    for nv in [1, 2, 4]:
        for nb in [1, 4, 8, 12]:
            rlda_inner_test(ns, min(nb, n_components), nb, nv)


def test_information():
    # test clustering + information
    ns = 6
    n_components = 3
    nb = 8
    nc = 2**nb
    n_model = 2000
    n_test = 2000
    noise = 30
    # Generate profiling data
    m = np.ones((nc, ns), dtype=np.int16)
    m[:, 0] = np.arange(nc)
    traces = np.random.randint(-noise, noise, (n_model, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, (n_model,), dtype=np.uint64)
    traces += m[labels].astype(np.int16)
    # Generate test data
    test_traces = np.random.randint(-noise, noise, (n_test, ns), dtype=np.int16)
    test_labels = np.random.randint(0, nc, n_test, dtype=np.uint16)
    test_traces += m[test_labels]

    rlda = RLDAClassifier(nb, n_components)
    rlda.fit_u(traces, labels[:, np.newaxis], 1)
    rlda.solve()
    prs = rlda.predict_proba(test_traces, 0)

    cl = rlda.get_clustered_model(0, 0, 2**16, False)
    it = RLDAInformationEstimator(cl, 0)
    # pi bounds should be exact bounds as t=0
    it.fit_u(test_traces, test_labels.astype(np.uint64))
    pi = it.get_information()
    pi_ref = nb + np.log2(prs[np.arange(n_test), test_labels]).mean()
    assert np.allclose(pi, pi_ref)


def test_rlda_fail_empty_classes():
    ns = 6
    n_components = 3
    nb = 8
    nc = 2**nb
    n_model = 2000
    noise = 30
    # Generate profiling data
    m = np.ones((nc, ns), dtype=np.int16)
    m[:, 0] = np.arange(nc)
    traces = np.random.randint(-noise, noise, (n_model, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, (n_model,), dtype=np.uint64)
    traces += m[labels].astype(np.int16)

    # make sure some bits have always the same value, here zero
    labels &= (nc >> 2) - 1

    rlda = RLDAClassifier(nb, n_components)
    rlda.fit_u(traces, labels[:, np.newaxis], 1)

    with pytest.raises(ScalibError):
        rlda.solve()
